import typing as t

from .markdown import Markdown, MarkdownType, Renderable
from .serialize import dumps, loads


class Link(Markdown):
    __regex__: str = r'(!?)\[(.*(?!.*\[(?!.*\])))\]\((.+?) ?((?:".+?")?)\)'

    def __init__(
        self,
        face: MarkdownType,
        path: str,
        title: str | None = None,
        media: bool = False,
    ) -> None:
        self.face = face
        self.path = path
        self.title = title
        self.media = media
        super().__init__({})

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}"
            f"({self.face!r}, {self.path!r}, media={self.media!r}, title={self.title!r})"
        )

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, Link):
            return False
        return (
            self.face == other.face
            and self.path == other.path
            and self.title == other.title
            and self.media == other.media
        )

    def __hash__(self) -> int:
        return hash(self.dumps())

    @property
    def __children__(self) -> list[Renderable]:
        children: list[Renderable] = [self.face, self.path]
        if self.title is not None:
            children.append(self.title)
        return children

    def dumps(self, *args, **kwargs) -> str:
        media = "!" if self.media else ""
        face = dumps(self.face, *args, **kwargs).strip()
        if self.title is not None:
            path = f"{self.path} {self.title}"
        else:
            path = self.path
        return f"{media}[{face}]({path})"

    @classmethod
    def loads(  # type: ignore[override]  # pylint: disable=arguments-differ, too-many-arguments
        cls,
        ctx: dict[str, t.Any] | None,
        media: str,
        face: str,
        path: str,
        title: str | None,
    ) -> MarkdownType:
        if not title:
            title = None
        return cls(
            face=t.cast(list[Markdown | str], loads(face))[0],
            path=path,
            title=title,
            media=bool(media),
        )



from .markdown import (
    Bold,
    CodeBlock,
    Header,
    InlineCode,
    Italics,
    Link,
    Markdown,
    Newline,
    Strikethrough,
    Table,
    Text,
)
from .markdown import __all__ as _markdown_all
from .markdown import dump, dumps, indexed_renderable, load, loads
from .search import find, findall, finditer

__all__ = ("find", "findall", "finditer") + _markdown_all



import typing as t

from .markdown import Markdown, MarkdownType, Renderable


class Header(Markdown):
    __regex__: str = r"\#+\ +(.+)\n((?:.|\n)+)"

    def __init__(
        self,
        name: MarkdownType,
        body: Renderable = "",
    ) -> None:
        self.name = name
        self.body = body
        super().__init__({})

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self.name!r}, {self.body!r})"

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, Header):
            return False
        return self.name == other.name and self.body == other.body

    def __hash__(self) -> int:
        return hash(self.dumps())

    @property
    def __children__(self) -> list[Renderable]:
        return [self.body]

    def dumps(self, *args: t.Any, depth: int = 0, **kwargs: t.Any) -> str:
        from .serialize import dumps  # pylint: disable=import-outside-toplevel

        body = dumps(self.body, *args, **kwargs, depth=depth + 1)
        return f"#{'#' * depth} {dumps(self.name, *args, **kwargs)}{body}"

    @classmethod
    def loads(  # type: ignore[override]  # pylint: disable=arguments-differ
        cls,
        _: dict[str, t.Any] | None,
        name: str,
        body: str,
    ) -> MarkdownType:
        # Prevents a circular import!
        from .serialize import loads  # pylint: disable=import-outside-toplevel

        return cls(name=name.strip(), body=loads(body))



import typing as t
from collections import defaultdict


class CaseInsensitiveDict(dict):
    @staticmethod
    def _scrub_key(key: str) -> str:
        return key.lower().replace("-", "_")

    def __setitem__(self, key: t.Any, value: t.Any) -> None:
        if not isinstance(key, str):
            raise TypeError("Key must be a string")
        super().__setitem__(self._scrub_key(key), value)

    def __getitem__(self, key: t.Any) -> t.Any:
        if not isinstance(key, str):
            raise TypeError("Key must be a string")
        return super().__getitem__(self._scrub_key(key))

    def __delitem__(self, key: t.Any) -> None:
        if not isinstance(key, str):
            raise TypeError("Key must be a string")
        return super().__delitem__(self._scrub_key(key))

    def __contains__(self, key: t.Any) -> bool:
        if not isinstance(key, str):
            return False
        return super().__contains__(self._scrub_key(key))


class Context:
    def __init__(self) -> None:
        self._storage: dict[Markdown, dict[str, t.Any]] = defaultdict(
            CaseInsensitiveDict
        )

    def __get__(self, instance: "Markdown", _: t.Type["Markdown"]) -> dict[str, t.Any]:
        return self._storage[instance]

    def __set__(self, instance: "Markdown", ctx: dict[str, t.Any]) -> None:
        for key, value in ctx.items():
            self._storage[instance][key] = value

    def __delete__(self, instance: "Markdown") -> None:
        del self._storage[instance]


class Markdown:
    __ignore__: bool = False
    __regex__: str = r"((?:.|\n(?<!\n))+)"
    __ctx_regex__: str = r"(?:<!--(\{.+\})-->\n)?"

    ctx: dict[str, t.Any] = t.cast(dict[str, t.Any], Context())

    parent: t.Optional["Markdown"] = None
    previous_sibling: t.Optional["Markdown"] = None
    next_sibling: t.Optional["Markdown"] = None

    def __init__(self, ctx: dict[str, t.Any]) -> None:
        self.ctx = ctx

    @property
    def __children__(self) -> list["Renderable"]:
        return []

    def __contains__(self, other: t.Any) -> bool:
        return other in self.__children__

    def __hash__(self) -> int:
        return hash(self.dumps())

    @classmethod
    def classes(cls) -> list[t.Type["Markdown"]]:
        _classes = cls.__subclasses__()
        return [_cls for _cls in _classes if not _cls.__ignore__]

    def dumps(self, *_, **__) -> str:
        return ""

    @classmethod
    def loads(cls, ctx: dict[str, t.Any] | None, content: str) -> "MarkdownType":
        return Text.loads(ctx, content)


MarkdownType = t.Union[Markdown, str]
Renderable = t.Union[
    Markdown,
    str,
    list[Markdown | str],
    list[Markdown],
    list[str],
]  # type: ignore[misc]


class Newline(Markdown):
    __regex__: str = r"\n"

    def __init__(self) -> None:
        super().__init__({})

    @staticmethod
    def dumps(*_: t.Any, **__: t.Any) -> str:
        return "\n"

    @classmethod
    def loads(cls, _: dict[str, t.Any]) -> MarkdownType:  # type: ignore[override]  # pylint: disable=arguments-differ
        return cls()


class Text(Markdown):
    def __init__(
        self,
        *elements: Renderable,
    ) -> None:
        self.elements = [
            element if not isinstance(element, str) else element.strip()
            for element in elements
        ]
        super().__init__({})

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, (Text, str)):
            return False
        if isinstance(other, str):
            return other == self.dumps()
        return self.elements == other.elements

    def __hash__(self) -> int:
        return hash(self.dumps())

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({', '.join(map(repr, self.elements))})"

    @property
    def __children__(self) -> list["Renderable"]:
        return self.elements

    def dumps(self, *args: t.Any, **kwargs: t.Any) -> str:
        # Prevents circular import!
        from .serialize import dumps  # pylint: disable=import-outside-toplevel

        kwargs["inline"] = True
        return " ".join(
            [
                dumps(
                    element,
                    *args,
                    **kwargs,
                )
                for element in self.elements
            ]
        )

    @classmethod
    def loads(  # pylint: disable=arguments-differ
        cls,
        _: dict[str, t.Any] | None,
        *elements: "MarkdownType",
    ) -> "MarkdownType":
        return cls(*elements)



from .codeblock import CodeBlock
from .header import Header
from .links import Link
from .markdown import Markdown, Newline, Text
from .serialize import dump, dumps, indexed_renderable, load, loads
from .style import Bold, InlineCode, Italics, Strikethrough
from .tables import Table

__all__ = (
    "dumps",
    "loads",
    "dump",
    "load",
    "indexed_renderable",
    "CodeBlock",
    "Markdown",
    "Table",
    "Header",
    "Bold",
    "Italics",
    "InlineCode",
    "Strikethrough",
    "Text",
    "Newline",
    "Link",
)



import typing as t

from .markdown import Markdown


class CodeBlock(Markdown):
    __regex__: str = r"```((?:.*)?)\n((?:.|\n)*?)```"

    def __init__(
        self,
        content: str,
        language: str | None = None,
        element_id: str | None = None,
    ) -> None:
        self.language = language
        self.content = content.strip()
        if element_id is not None:
            super().__init__({"element-id": element_id})
        else:
            super().__init__({})

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, CodeBlock):
            return False
        return self.language == other.language and self.content == other.content

    def __hash__(self) -> int:
        return hash(self.dumps())

    def __repr__(self) -> str:
        params = [repr(param) for param in [self.language] if param]
        repr_params = ", ".join([repr(self.content)[:30] + "...", *params])
        return f"{self.__class__.__qualname__}({repr_params})"

    def dumps(self, *_: t.Any, **__: t.Any) -> str:
        return f"```{self.language if self.language else ''}\n{self.content}\n```\n"

    @classmethod
    def loads(  # type: ignore[override] # pylint: disable=arguments-differ
        cls,
        _: dict[str, t.Any] | None,
        language: str,
        content: str,
    ) -> Markdown:
        language = language.strip()
        return cls(content, language if language else None)



import json
import re
import typing as t
from collections import abc

from .codeblock import CodeBlock
from .markdown import Markdown, Renderable
from .parse import build_lexer, construct, scrub

CLEANING_IGNORE_PATTERNS = [
    r"<!--.*?-->",  # HTML comments
    CodeBlock.__regex__,  # Code blocks
]

CLEANER_PATTERNS = [
    (r"^ +(.*)$", r"\1", re.MULTILINE),
    (r"^(.*) +$", r"\1", re.MULTILINE),
    (r"\n{3}", r"\n\n", re.NOFLAG),
    (r" {2}", r" ", re.NOFLAG),
]


def _clean(rendered: str) -> str:
    while True:
        new_rendered = rendered
        for pattern, replacement, flags in CLEANER_PATTERNS:
            if match := re.search(pattern, new_rendered, flags=flags):
                if any(
                    ignore_match.start() <= match.start() <= ignore_match.end()
                    for ignore_pattern in CLEANING_IGNORE_PATTERNS
                    for ignore_match in re.finditer(
                        ignore_pattern, new_rendered, flags=flags
                    )
                ):
                    continue
                new_rendered = re.sub(pattern, replacement, new_rendered, flags=flags)
        if new_rendered != rendered:
            rendered = new_rendered
        else:
            break
    return rendered


def indexed_renderable(renderable: Renderable) -> Renderable:
    _asign_parents(renderable)
    _asign_siblings(renderable)
    return renderable


def _asign_parents(renderable: Renderable) -> None:
    if isinstance(renderable, Markdown):
        for child in renderable.__children__:
            if isinstance(child, Markdown):
                child.parent = renderable
                _asign_parents(child)
    elif isinstance(renderable, abc.Iterable):
        for child in renderable:
            _asign_parents(child)


def _asign_siblings(renderable: Renderable) -> None:
    if isinstance(renderable, Markdown):
        for child in renderable.__children__:
            _asign_siblings(child)
    elif isinstance(renderable, abc.Iterable):
        prev = None
        for child in renderable:
            if prev is not None:
                prev.next_sibling = child
                child.previous_sibling = prev
            prev = child


def dumps(
    content: Renderable,
    *args: t.Any,
    inline: bool = False,
    clean: bool = True,
    **kwargs: t.Any,
) -> str:
    result = ""
    if isinstance(content, Markdown):  # pylint: disable=unidiomatic-typecheck
        if content.ctx:
            result += f"\n\n<!--{json.dumps(content.ctx)}-->"
        result += content.dumps(*args, **kwargs, clean=False)
    elif isinstance(content, str):
        result += content
        if inline is False:
            result += "\n\n"
    elif isinstance(content, abc.Iterable):
        for item in content:
            result += dumps(item, *args, **kwargs, clean=False)
    if clean:
        return _clean(result)
    return result


def loads(content: str) -> Renderable:
    if content == "":
        return [content]
    lexer = build_lexer()
    lexer.input(content)  # type: ignore[no-untyped-call]
    lookup = {cls.__qualname__.upper(): cls for cls in Markdown.classes()}
    return indexed_renderable(
        scrub([construct(token.value, lookup[token.type]) for token in lexer])
    )


def dump(content: Renderable, file: t.TextIO, *args: t.Any, **kwargs: t.Any) -> None:
    with file:
        file.write(dumps(content, *args, **kwargs))


def load(file: t.TextIO) -> Renderable:
    with file:
        return loads(file.read())



import re
import typing as t

from .markdown import Markdown, MarkdownType, Renderable
from .serialize import dumps, loads


def pad(element: Renderable, *args: t.Any, **kwargs: t.Any) -> str:
    return f"  {dumps(element, *args, **kwargs, inline=True)}  "


def match_alignment(alignment: str) -> str:
    if re.search(r":-+:", alignment):
        return "center"
    if re.search(r":-+", alignment):
        return "right"
    if re.search(r"-+:", alignment):
        return "left"
    return "default"


def extract(
    raw_headers: str,
    raw_alignment: str,
    raw_rows: str,
) -> tuple[list[list[MarkdownType]], list[MarkdownType], list[str]]:
    _, *columns, _ = map(str.strip, raw_headers.split("|"))
    _, *alignment, _ = map(str.strip, raw_alignment.split("|"))
    rows = []
    for row in raw_rows.splitlines():
        _, *elements, _ = map(str.strip, row.split("|"))
        rows.append(
            list(
                map(
                    lambda element: t.cast(list[Markdown | str], loads(element))[0],
                    elements,
                )
            )
        )
    return (
        rows,
        list(
            map(lambda column: t.cast(list[Markdown | str], loads(column))[0], columns)
        ),
        list(map(match_alignment, alignment)),
    )


class Table(Markdown):
    __regex__: str = r"(\|(?:.+\|)+)\n\|(:?-+:?\|)+\n((?:\|(?:.+\|)+\n)+)"

    def __init__(
        self,
        rows: t.Sequence[t.Sequence[Renderable]],
        columns: t.Sequence[Renderable],
        alignment: t.Sequence[str] = (),
        element_id: str | None = None,
    ) -> None:
        self.columns = tuple(columns)
        self.rows = tuple(tuple(row) for row in rows)
        self.alignment = alignment if alignment else ("default",) * len(columns)
        self._element_id = element_id
        if element_id is not None:
            super().__init__({"element-id": element_id})
        else:
            super().__init__({})

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__} columns={self.columns!r} entries={self.height - 1}>"

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, Table):
            return False
        return (
            self.columns == other.columns
            and self.rows == other.rows
            and self.element_id == other.element_id
        )

    def __hash__(self) -> int:
        return hash(self.dumps())

    @property
    def width(self) -> int:
        return len(self.columns)

    @property
    def height(self) -> int:
        return len(self.rows) + 1

    @property
    def element_id(self) -> str | None:
        return self._element_id

    def dumps(
        self,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> str:
        headers = [pad(header, *args, **kwargs) for header in self.columns]
        alignment = [
            (":" if alignment in ("left", "center") else "")
            + "-" * len(item)
            + (":" if alignment in ("right", "center") else "")
            for item, alignment in zip(headers, self.alignment)
        ]
        table = [
            f"|{'|'.join(headers)}|",
            f"|{'|'.join(alignment)}|",
        ] + [
            f"|{'|'.join([pad(item, *args, **kwargs) for item in row])}|"
            for row in self.rows
        ]
        return "\n" + "\n".join(table) + "\n\n"

    @classmethod
    def loads(  # type: ignore[override]  # pylint: disable=arguments-differ
        cls,
        ctx: t.Optional[dict[str, t.Any]],
        raw_headers: str,
        raw_alignment: str,
        raw_rows: str,
    ) -> MarkdownType:
        element_id = ctx.get("element-id") if ctx is not None else None
        rows, columns, alignment = extract(raw_headers, raw_alignment, raw_rows)
        return cls(rows, columns, alignment, element_id=element_id)



from collections import abc
from typing import Any, Generator, Type, cast

from .markdown.markdown import Markdown, MarkdownType, Renderable


def _match_query(
    markdown: Markdown,
    tag: Type[Markdown] | None,
    query: MarkdownType | None,
    **ctx: Any,
) -> bool:
    conditions = []
    if tag is not None:
        conditions.append(
            type(markdown) == tag  # pylint: disable=unidiomatic-typecheck
        )
    if query is not None:
        conditions.append(query in markdown)
    if ctx:
        conditions.append(
            all(
                markdown.ctx[key] == value if key in markdown.ctx else False
                for key, value in ctx.items()
            )
        )
    return all(conditions)


def finditer(
    markdown: Renderable | list[Renderable],
    tag: Type[Markdown] | None = None,
    query: MarkdownType | None = None,
    parent: Renderable | None = None,
    **ctx: Any,
) -> Generator[Renderable, None, None]:
    if isinstance(markdown, Markdown):
        if _match_query(markdown, tag=tag, query=query, **ctx):
            yield markdown
        if markdown.__children__:
            yield from finditer(
                markdown.__children__, tag=tag, query=query, **ctx, parent=markdown
            )
    elif isinstance(markdown, str):
        if isinstance(query, str):
            if query in markdown:
                if parent is not None:
                    yield parent
                else:
                    yield markdown
    elif isinstance(markdown, abc.Iterable):
        for item in markdown:
            yield from finditer(
                item, tag=tag, query=query, **ctx, parent=cast(Renderable, parent)
            )


def findall(
    markdown: Renderable,
    tag: Type[Markdown] | None = None,
    query: MarkdownType | None = None,
    **ctx: Any,
) -> list[Renderable]:
    return list(finditer(markdown, tag=tag, query=query, **ctx))


def find(
    markdown: Renderable,
    tag: Type[Markdown] | None = None,
    query: MarkdownType | None = None,
    **ctx: Any,
) -> Renderable | None:
    for item in finditer(markdown, tag=tag, query=query, **ctx):
        return item
    return None



import typing as t

from .markdown import MarkdownType, Renderable, Text


class StyleMixin(Text):
    __ignore__: bool = True

    def __init__(self, content: str) -> None:
        super().__init__(content.strip())

    @property
    def content(self) -> str:
        return t.cast(str, self.elements[0])

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self.content!r})"

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self.content == other.content

    def __hash__(self) -> int:
        return hash(self.dumps())

    @property
    def __children__(self) -> list[Renderable]:
        return [self.content]

    @classmethod
    def loads(cls, _: dict[str, t.Any] | None, content: str) -> MarkdownType:  # type: ignore[override] # pylint: disable=arguments-differ
        return cls(content)

    def dumps(self, *args: t.Any, **kwargs: t.Any) -> str:
        # Prevents circular import!
        from .serialize import dumps  # pylint: disable=import-outside-toplevel

        return dumps(self.content.strip(), *args, **kwargs).strip()


class Bold(StyleMixin):
    __regex__: str = r"(?<!\\)\*\*(.+?)(?<!\\)\*\*"

    def dumps(self, *args: t.Any, **kwargs: t.Any) -> str:
        return f" **{super().dumps(*args, **kwargs)}** "


class Italics(StyleMixin):
    __regex__: str = r"(?<!\\)\*(.+?)(?<!\\)\*"

    def dumps(self, *args: t.Any, **kwargs: t.Any) -> str:
        return f" *{super().dumps(*args, **kwargs)}* "


class Strikethrough(StyleMixin):
    __regex__: str = r"(?<!\\)~~(.+?)(?<!\\)~~"

    def dumps(self, *args: t.Any, **kwargs: t.Any) -> str:
        return f" ~~{super().dumps(*args, **kwargs)}~~ "


class InlineCode(StyleMixin):
    __regex__: str = r"(?:(?<!\\)``(?!`)(.+?)(?<!\\)``)|(?:(?<!\\)`(?!`)\1(?<!\\)`)"

    def dumps(self, *args: t.Any, **kwargs: t.Any) -> str:
        return f" ``{super().dumps(*args, **kwargs)}`` "



import json
import re
import typing as t

from .lex import Lexer, LexToken, lex
from .markdown import Markdown, MarkdownType, Newline, Text
from .style import StyleMixin


def tokens() -> dict[str, t.Type[Markdown]]:
    return {cls.__qualname__.upper(): cls for cls in Markdown.classes()}


def token_regex() -> t.Generator[tuple[str, str], None, None]:
    for name, cls in tokens().items():
        yield f"t_{name}", cls.__ctx_regex__ + cls.__regex__


def lex_error(token: LexToken) -> None:
    raise ValueError(token)


def build_lexer() -> Lexer:
    order_preserved_attributes = (
        ("tokens", tuple(tokens())),
        ("t_error", lex_error),
        *token_regex(),
    )
    cls = type("LemonLexer", (), {})
    for name, value in order_preserved_attributes:
        setattr(cls, name, value)
    return t.cast(Lexer, lex(module=cls))  # type: ignore[no-untyped-call]


def extract(value: str, cls: t.Type[Markdown]) -> tuple[str | None, list[str]]:
    match = re.fullmatch(
        cls.__ctx_regex__ + cls.__regex__,
        value,
    )
    assert match is not None
    ctx, *params = t.cast(list[str], match.groups())
    return t.cast(str | None, ctx), params


def subdivide(markdown: Text) -> Text:
    source, *_ = markdown.elements
    assert isinstance(source, str)
    elements: list[MarkdownType] = []
    try:
        (match, style_class), *_ = sorted(
            filter(
                lambda pair: pair[0] is not None,
                [
                    (
                        t.cast(t.Match[str], re.search(style_class.__regex__, source)),
                        style_class,
                    )
                    for style_class in StyleMixin.__subclasses__()
                ],
            ),
            key=lambda pair: pair[0].start(),
        )
        sub_content = match.string[match.start() : match.end()]
        pre, source = source.split(sub_content)
        if pre.strip():
            elements += [pre]
        elements += [
            construct(sub_content, style_class),
            *subdivide(Text(source)).elements,  # type: ignore[list-item]
        ]
    except ValueError:
        if source.strip():
            elements.append(source.strip())
    return Text(*elements)


def construct(value: str, cls: t.Type[Markdown]) -> MarkdownType:
    ctx, params = extract(value, cls)
    if ctx is not None:
        element = cls.loads(
            t.cast(
                dict[str, t.Any] | None,
                json.loads(ctx),
            ),
            *params,
        )
    else:
        element = cls.loads(ctx, *params)
    # Check only for Text objects
    if type(element) == Text:  # pylint: disable=unidiomatic-typecheck
        return subdivide(element)
    return element


def scrub(tree: list[MarkdownType]) -> list[MarkdownType]:
    tree = [
        element
        for element in tree
        if not (isinstance(element, str) and element == " " * len(element))
    ]
    tree = [element for element in tree if not isinstance(element, Newline)]
    return tree
